/* -*- c-file-style: "java"; indent-tabs-mode: nil -*- */

#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <sys/select.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/time.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <avahi-common/simple-watch.h>
#include <avahi-common/error.h>
#include <avahi-common/alternative.h>
#include <avahi-common/malloc.h>
#include <avahi-client/publish.h>

#include "distcc.h"
#include "zeroconf.h"
#include "trace.h"
#include "exitcode.h"

struct context {
    int thread_running;
    pthread_t thread_id;
    pthread_mutex_t mutex;
    char *name;
    AvahiSimplePoll *simple_poll;
    AvahiClient *client;
    AvahiEntryGroup *group;
    uint16_t port;
    int n_cpus;
};

static void publish_reply(AvahiEntryGroup *g, AvahiEntryGroupState state, void *userdata);

static void register_stuff(struct context *ctx) {

    if (!ctx->group) {

        if (!(ctx->group = avahi_entry_group_new(ctx->client, publish_reply, ctx))) {
            rs_log_crit("Failed to create entry group: %s\n", avahi_strerror(avahi_client_errno(ctx->client)));
            goto fail;
        }
        
    }

    if (avahi_entry_group_is_empty(ctx->group)) {
        char cpus[32];

        snprintf(cpus, sizeof(cpus), "cpus=%i", ctx->n_cpus);

        /* Register our service */
        
        if (avahi_entry_group_add_service(ctx->group,
                                          AVAHI_IF_UNSPEC,
                                          AVAHI_PROTO_UNSPEC,
                                          0,
                                          ctx->name,
                                          DCC_DNS_SERVICE_TYPE,
                                          NULL,
                                          NULL,
                                          ctx->port,
                                          "txtvers=1",
                                          cpus,
                                          "distcc="PACKAGE_VERSION,
                                          "gnuhost="GNU_HOST,
                                          NULL) < 0) {
            rs_log_crit("Failed to add service: %s\n", avahi_strerror(avahi_client_errno(ctx->client)));
            goto fail;
        }

        if (avahi_entry_group_commit(ctx->group) < 0) {
            rs_log_crit("Failed to commit entry group: %s\n", avahi_strerror(avahi_client_errno(ctx->client)));
            goto fail;
        }
    }

    return;
    
    fail:
    avahi_simple_poll_quit(ctx->simple_poll);
}

/* Called when publishing of service data completes */
static void publish_reply(AvahiEntryGroup *UNUSED(g), AvahiEntryGroupState state, void *userdata) {
    struct context *ctx = userdata;

    switch (state) {
        
    case AVAHI_ENTRY_GROUP_COLLISION: {
        char *n;

        /* Pick a new name for our service */
        
        n = avahi_alternative_service_name(ctx->name);
        assert(n);
        
        avahi_free(ctx->name);
        ctx->name = n;

        register_stuff(ctx);
        break;
    }

    case AVAHI_ENTRY_GROUP_FAILURE: 
        rs_log_crit("Failed to register service: %s\n", avahi_strerror(avahi_client_errno(ctx->client)));
        avahi_simple_poll_quit(ctx->simple_poll);
        break;
        
    case AVAHI_ENTRY_GROUP_UNCOMMITED:
    case AVAHI_ENTRY_GROUP_REGISTERING:
    case AVAHI_ENTRY_GROUP_ESTABLISHED:
        ;
    }
}

static void client_callback(AvahiClient *client, AvahiClientState state, void *userdata) {
    struct context *ctx = userdata;

    ctx->client = client;
    
    switch (state) {

    case AVAHI_CLIENT_S_RUNNING:

        register_stuff(ctx);
        break;

    case AVAHI_CLIENT_S_COLLISION:

        if (ctx->group)
            avahi_entry_group_reset(ctx->group);
        break;

    case AVAHI_CLIENT_FAILURE:

        if (avahi_client_errno(client) == AVAHI_ERR_DISCONNECTED) {
            int error;

            avahi_client_free(ctx->client);
            ctx->client = NULL;
            ctx->group = NULL;

            /* Reconnect to the server */

            if (!(ctx->client = avahi_client_new(avahi_simple_poll_get(ctx->simple_poll),
                                                 AVAHI_CLIENT_NO_FAIL,
                                                 client_callback,
                                                 ctx,
                                                 &error))) {
                
                rs_log_crit("Failed to contact server: %s\n", avahi_strerror(error));
                avahi_simple_poll_quit(ctx->simple_poll);
            }
            
        } else {
            rs_log_crit("Client failure: %s\n", avahi_strerror(avahi_client_errno(client)));
            avahi_simple_poll_quit(ctx->simple_poll);
        }

        break;
        
    case AVAHI_CLIENT_S_REGISTERING:
    case AVAHI_CLIENT_CONNECTING:
        ;
    }
}

static void* thread(void *userdata) {
    struct context *ctx = userdata;
    sigset_t mask;
    int r;

    /* Make sure that signals are delivered to the main thread */
    sigfillset(&mask);
    pthread_sigmask(SIG_BLOCK, &mask, NULL);
    
    pthread_mutex_lock(&ctx->mutex);

    /* Run the main loop */
    r = avahi_simple_poll_loop(ctx->simple_poll);

    /* Cleanup some stuff */
    if (ctx->client)
        avahi_client_free(ctx->client);
    ctx->client = NULL;
    ctx->group = NULL;
    
    pthread_mutex_unlock(&ctx->mutex);
    
    return NULL;
}

static int poll_func(struct pollfd *ufds, unsigned int nfds, int timeout, void *userdata) {
    pthread_mutex_t *mutex = userdata;
    int r;

    /* Before entering poll() we unlock the mutex, so that
     * avahi_simple_poll_quit() can succeed from another thread. */
    
    pthread_mutex_unlock(mutex);
    r = poll(ufds, nfds, timeout);
    pthread_mutex_lock(mutex);

    return r;
}

/* register a distcc service in DNS-SD/mDNS with the given port and number of CPUs */
void* dcc_zeroconf_register(uint16_t port, int n_cpus) {
    struct context *ctx = NULL;
    
    char service[256] = "distcc@";
    int error, ret;

    ctx = malloc(sizeof(struct context));
    assert(ctx);
    ctx->client = NULL;
    ctx->group = NULL;
    ctx->simple_poll = NULL;
    ctx->thread_running = 0;
    ctx->port = port;
    ctx->n_cpus = n_cpus;
    pthread_mutex_init(&ctx->mutex, NULL);
    
    /* Prepare service name */
    gethostname(service+7, sizeof(service)-8);
    service[sizeof(service)-1] = 0;

    ctx->name = strdup(service);
    assert(ctx->name);

    if (!(ctx->simple_poll = avahi_simple_poll_new())) {
        rs_log_crit("Failed to create event loop object.\n");
        goto fail;
    }

    avahi_simple_poll_set_func(ctx->simple_poll, poll_func, &ctx->mutex);
    
    if (!(ctx->client = avahi_client_new(avahi_simple_poll_get(ctx->simple_poll), AVAHI_CLIENT_NO_FAIL, client_callback, ctx, &error))) {
        rs_log_crit("Failed to create client object: %s\n", avahi_strerror(avahi_client_errno(ctx->client)));
        goto fail;
    }

    /* Create the mDNS event handler */
    if ((ret = pthread_create(&ctx->thread_id, NULL, thread, ctx)) < 0) {
        rs_log_crit("Failed to create thread: %s\n", strerror(ret));
        goto fail;
    }

    ctx->thread_running = 1;
    
    return ctx;
    
 fail:

    if (ctx)
        dcc_zeroconf_unregister(ctx);
    
    return NULL;
}

/* Unregister this server from DNS-SD/mDNS */
int dcc_zeroconf_unregister(void *u) {
    struct context *ctx = u;

    if (ctx->thread_running) {
        pthread_mutex_lock(&ctx->mutex);
        avahi_simple_poll_quit(ctx->simple_poll);
        pthread_mutex_unlock(&ctx->mutex);
        
        pthread_join(ctx->thread_id, NULL);
        ctx->thread_running = 0;
    }

    avahi_free(ctx->name);

    if (ctx->client)
        avahi_client_free(ctx->client);

    if (ctx->simple_poll)
        avahi_simple_poll_free(ctx->simple_poll);

    pthread_mutex_destroy(&ctx->mutex);

    free(ctx);
    
    return 0;
}
